import { isAbstractType } from "../type/definition.mjs";

import { GraphQLSkipDirective, GraphQLIncludeDirective } from "../type/directives.mjs";

import { typeFromAST } from "../utilities/typeFromAST.mjs";

import { getDirectiveValues } from "./values.mjs";

import { Kind } from "@0no-co/graphql.web";

function collectFields(schema, fragments, variableValues, runtimeType, selectionSet) {
  const fields = new Map;
  collectFieldsImpl(schema, fragments, variableValues, runtimeType, selectionSet, fields, new Set);
  return fields;
}

function collectSubfields(schema, fragments, variableValues, returnType, fieldNodes) {
  const subFieldNodes = new Map;
  const visitedFragmentNames = new Set;
  for (const node of fieldNodes) {
    if (node.selectionSet) {
      collectFieldsImpl(schema, fragments, variableValues, returnType, node.selectionSet, subFieldNodes, visitedFragmentNames);
    }
  }
  return subFieldNodes;
}

function collectFieldsImpl(schema, fragments, variableValues, runtimeType, selectionSet, fields, visitedFragmentNames) {
  for (const selection of selectionSet.selections) {
    switch (selection.kind) {
     case Kind.FIELD:
      {
        if (!shouldIncludeNode(variableValues, selection)) {
          continue;
        }
        const name = (node = selection).alias ? node.alias.value : node.name.value;
        const fieldList = fields.get(name);
        if (void 0 !== fieldList) {
          fieldList.push(selection);
        } else {
          fields.set(name, [ selection ]);
        }
        break;
      }

     case Kind.INLINE_FRAGMENT:
      if (!shouldIncludeNode(variableValues, selection) || !doesFragmentConditionMatch(schema, selection, runtimeType)) {
        continue;
      }
      collectFieldsImpl(schema, fragments, variableValues, runtimeType, selection.selectionSet, fields, visitedFragmentNames);
      break;

     case Kind.FRAGMENT_SPREAD:
      {
        const fragName = selection.name.value;
        if (visitedFragmentNames.has(fragName) || !shouldIncludeNode(variableValues, selection)) {
          continue;
        }
        visitedFragmentNames.add(fragName);
        const fragment = fragments[fragName];
        if (!fragment || !doesFragmentConditionMatch(schema, fragment, runtimeType)) {
          continue;
        }
        collectFieldsImpl(schema, fragments, variableValues, runtimeType, fragment.selectionSet, fields, visitedFragmentNames);
        break;
      }
    }
  }
  var node;
}

function shouldIncludeNode(variableValues, node) {
  const skip = getDirectiveValues(GraphQLSkipDirective, node, variableValues);
  if (!0 === (null == skip ? void 0 : skip.if)) {
    return !1;
  }
  const include = getDirectiveValues(GraphQLIncludeDirective, node, variableValues);
  if (!1 === (null == include ? void 0 : include.if)) {
    return !1;
  }
  return !0;
}

function doesFragmentConditionMatch(schema, fragment, type) {
  const typeConditionNode = fragment.typeCondition;
  if (!typeConditionNode) {
    return !0;
  }
  const conditionalType = typeFromAST(schema, typeConditionNode);
  if (conditionalType === type) {
    return !0;
  }
  if (isAbstractType(conditionalType)) {
    return schema.isSubType(conditionalType, type);
  }
  return !1;
}

export { collectFields, collectSubfields };
