var jsutils_inspect = require("../jsutils/inspect.js");

var jsutils_invariant = require("../jsutils/invariant.js");

var jsutils_isIterableObject = require("../jsutils/isIterableObject.js");

var jsutils_isObjectLike = require("../jsutils/isObjectLike.js");

var jsutils_isPromise = require("../jsutils/isPromise.js");

var jsutils_memoize3 = require("../jsutils/memoize3.js");

var jsutils_Path = require("../jsutils/Path.js");

var jsutils_promiseForObject = require("../jsutils/promiseForObject.js");

var jsutils_promiseReduce = require("../jsutils/promiseReduce.js");

var error_locatedError = require("../error/locatedError.js");

var language_ast = require("../language/ast.js");

var type_definition = require("../type/definition.js");

var type_introspection = require("../type/introspection.js");

var type_validate = require("../type/validate.js");

var execution_collectFields = require("./collectFields.js");

var execution_values = require("./values.js");

var graphql_web = require("@0no-co/graphql.web");

const collectSubfields = jsutils_memoize3.memoize3(((exeContext, returnType, fieldNodes) => execution_collectFields.collectSubfields(exeContext.schema, exeContext.fragments, exeContext.variableValues, returnType, fieldNodes)));

function execute(args) {
  const {schema, rootValue} = args;
  assertValidExecutionArguments(schema);
  const exeContext = buildExecutionContext(args);
  if (!("schema" in exeContext)) {
    return {
      errors: exeContext
    };
  }
  try {
    const {operation} = exeContext;
    const result = function executeOperation(exeContext, operation, rootValue) {
      const rootType = exeContext.schema.getRootType(operation.operation);
      if (null == rootType) {
        throw new graphql_web.GraphQLError(`Schema is not configured to execute ${operation.operation} operation.`, {
          nodes: operation
        });
      }
      const rootFields = execution_collectFields.collectFields(exeContext.schema, exeContext.fragments, exeContext.variableValues, rootType, operation.selectionSet);
      const path = void 0;
      switch (operation.operation) {
       case language_ast.OperationTypeNode.QUERY:
        return executeFields(exeContext, rootType, rootValue, path, rootFields);

       case language_ast.OperationTypeNode.MUTATION:
        return function executeFieldsSerially(exeContext, parentType, sourceValue, path, fields) {
          return jsutils_promiseReduce.promiseReduce(fields.entries(), ((results, [responseName, fieldNodes]) => {
            const fieldPath = jsutils_Path.addPath(path, responseName, parentType.name);
            const result = executeField(exeContext, parentType, sourceValue, fieldNodes, fieldPath);
            if (void 0 === result) {
              return results;
            }
            if (jsutils_isPromise.isPromise(result)) {
              return result.then((resolvedResult => {
                results[responseName] = resolvedResult;
                return results;
              }));
            }
            results[responseName] = result;
            return results;
          }), Object.create(null));
        }(exeContext, rootType, rootValue, path, rootFields);

       case language_ast.OperationTypeNode.SUBSCRIPTION:
        return executeFields(exeContext, rootType, rootValue, path, rootFields);
      }
    }(exeContext, operation, rootValue);
    if (jsutils_isPromise.isPromise(result)) {
      return result.then((data => buildResponse(data, exeContext.errors)), (error => {
        exeContext.errors.push(error);
        return buildResponse(null, exeContext.errors);
      }));
    }
    return buildResponse(result, exeContext.errors);
  } catch (error) {
    exeContext.errors.push(error);
    return buildResponse(null, exeContext.errors);
  }
}

function buildResponse(data, errors) {
  return 0 === errors.length ? {
    data
  } : {
    errors,
    data
  };
}

function assertValidExecutionArguments(schema, document, rawVariableValues) {
  type_validate.assertValidSchema(schema);
}

function buildExecutionContext(args) {
  var _definition$name, _operation$variableDe;
  const {schema, document, rootValue, contextValue, variableValues: rawVariableValues, operationName, fieldResolver, typeResolver, subscribeFieldResolver} = args;
  let operation;
  const fragments = Object.create(null);
  for (const definition of document.definitions) {
    switch (definition.kind) {
     case graphql_web.Kind.OPERATION_DEFINITION:
      if (null == operationName) {
        if (void 0 !== operation) {
          return [ new graphql_web.GraphQLError("Must provide operation name if query contains multiple operations.") ];
        }
        operation = definition;
      } else if ((null === (_definition$name = definition.name) || void 0 === _definition$name ? void 0 : _definition$name.value) === operationName) {
        operation = definition;
      }
      break;

     case graphql_web.Kind.FRAGMENT_DEFINITION:
      fragments[definition.name.value] = definition;
    }
  }
  if (!operation) {
    if (null != operationName) {
      return [ new graphql_web.GraphQLError(`Unknown operation named "${operationName}".`) ];
    }
    return [ new graphql_web.GraphQLError("Must provide an operation.") ];
  }
  const variableDefinitions = null !== (_operation$variableDe = operation.variableDefinitions) && void 0 !== _operation$variableDe ? _operation$variableDe : [];
  const coercedVariableValues = execution_values.getVariableValues(schema, variableDefinitions, null != rawVariableValues ? rawVariableValues : {}, {
    maxErrors: 50
  });
  if (coercedVariableValues.errors) {
    return coercedVariableValues.errors;
  }
  return {
    schema,
    fragments,
    rootValue,
    contextValue,
    operation,
    variableValues: coercedVariableValues.coerced,
    fieldResolver: null != fieldResolver ? fieldResolver : defaultFieldResolver,
    typeResolver: null != typeResolver ? typeResolver : defaultTypeResolver,
    subscribeFieldResolver: null != subscribeFieldResolver ? subscribeFieldResolver : defaultFieldResolver,
    errors: []
  };
}

function executeFields(exeContext, parentType, sourceValue, path, fields) {
  const results = Object.create(null);
  let containsPromise = !1;
  for (const [responseName, fieldNodes] of fields.entries()) {
    const result = executeField(exeContext, parentType, sourceValue, fieldNodes, jsutils_Path.addPath(path, responseName, parentType.name));
    if (void 0 !== result) {
      results[responseName] = result;
      if (jsutils_isPromise.isPromise(result)) {
        containsPromise = !0;
      }
    }
  }
  if (!containsPromise) {
    return results;
  }
  return jsutils_promiseForObject.promiseForObject(results);
}

function executeField(exeContext, parentType, source, fieldNodes, path) {
  var _fieldDef$resolve;
  const fieldDef = getFieldDef(exeContext.schema, parentType, fieldNodes[0]);
  if (!fieldDef) {
    return;
  }
  const returnType = fieldDef.type;
  const resolveFn = null !== (_fieldDef$resolve = fieldDef.resolve) && void 0 !== _fieldDef$resolve ? _fieldDef$resolve : exeContext.fieldResolver;
  const info = buildResolveInfo(exeContext, fieldDef, fieldNodes, parentType, path);
  try {
    const args = execution_values.getArgumentValues(fieldDef, fieldNodes[0], exeContext.variableValues);
    const result = resolveFn(source, args, exeContext.contextValue, info);
    let completed;
    if (jsutils_isPromise.isPromise(result)) {
      completed = result.then((resolved => completeValue(exeContext, returnType, fieldNodes, info, path, resolved)));
    } else {
      completed = completeValue(exeContext, returnType, fieldNodes, info, path, result);
    }
    if (jsutils_isPromise.isPromise(completed)) {
      return completed.then(void 0, (rawError => handleFieldError(error_locatedError.locatedError(rawError, fieldNodes, jsutils_Path.pathToArray(path)), returnType, exeContext)));
    }
    return completed;
  } catch (rawError) {
    return handleFieldError(error_locatedError.locatedError(rawError, fieldNodes, jsutils_Path.pathToArray(path)), returnType, exeContext);
  }
}

function buildResolveInfo(exeContext, fieldDef, fieldNodes, parentType, path) {
  return {
    fieldName: fieldDef.name,
    fieldNodes,
    returnType: fieldDef.type,
    parentType,
    path,
    schema: exeContext.schema,
    fragments: exeContext.fragments,
    rootValue: exeContext.rootValue,
    operation: exeContext.operation,
    variableValues: exeContext.variableValues
  };
}

function handleFieldError(error, returnType, exeContext) {
  if (type_definition.isNonNullType(returnType)) {
    throw error;
  }
  exeContext.errors.push(error);
  return null;
}

function completeValue(exeContext, returnType, fieldNodes, info, path, result) {
  if (result instanceof Error) {
    throw result;
  }
  if (type_definition.isNonNullType(returnType)) {
    const completed = completeValue(exeContext, returnType.ofType, fieldNodes, info, path, result);
    if (null === completed) {
      throw new Error(`Cannot return null for non-nullable field ${info.parentType.name}.${info.fieldName}.`);
    }
    return completed;
  }
  if (null == result) {
    return null;
  }
  if (type_definition.isListType(returnType)) {
    return function completeListValue(exeContext, returnType, fieldNodes, info, path, result) {
      if (!jsutils_isIterableObject.isIterableObject(result)) {
        throw new graphql_web.GraphQLError(`Expected Iterable, but did not find one for field "${info.parentType.name}.${info.fieldName}".`);
      }
      const itemType = returnType.ofType;
      let containsPromise = !1;
      const completedResults = Array.from(result, ((item, index) => {
        const itemPath = jsutils_Path.addPath(path, index, void 0);
        try {
          let completedItem;
          if (jsutils_isPromise.isPromise(item)) {
            completedItem = item.then((resolved => completeValue(exeContext, itemType, fieldNodes, info, itemPath, resolved)));
          } else {
            completedItem = completeValue(exeContext, itemType, fieldNodes, info, itemPath, item);
          }
          if (jsutils_isPromise.isPromise(completedItem)) {
            containsPromise = !0;
            return completedItem.then(void 0, (rawError => handleFieldError(error_locatedError.locatedError(rawError, fieldNodes, jsutils_Path.pathToArray(itemPath)), itemType, exeContext)));
          }
          return completedItem;
        } catch (rawError) {
          return handleFieldError(error_locatedError.locatedError(rawError, fieldNodes, jsutils_Path.pathToArray(itemPath)), itemType, exeContext);
        }
      }));
      return containsPromise ? Promise.all(completedResults) : completedResults;
    }(exeContext, returnType, fieldNodes, info, path, result);
  }
  if (type_definition.isLeafType(returnType)) {
    return function completeLeafValue(returnType, result) {
      const serializedResult = returnType.serialize(result);
      if (null == serializedResult) {
        throw new Error(`Expected \`${jsutils_inspect.inspect(returnType)}.serialize(${jsutils_inspect.inspect(result)})\` to return non-nullable value, returned: ${jsutils_inspect.inspect(serializedResult)}`);
      }
      return serializedResult;
    }(returnType, result);
  }
  if (type_definition.isAbstractType(returnType)) {
    return function completeAbstractValue(exeContext, returnType, fieldNodes, info, path, result) {
      var _returnType$resolveTy;
      const resolveTypeFn = null !== (_returnType$resolveTy = returnType.resolveType) && void 0 !== _returnType$resolveTy ? _returnType$resolveTy : exeContext.typeResolver;
      const contextValue = exeContext.contextValue;
      const runtimeType = resolveTypeFn(result, contextValue, info, returnType);
      if (jsutils_isPromise.isPromise(runtimeType)) {
        return runtimeType.then((resolvedRuntimeType => completeObjectValue(exeContext, ensureValidRuntimeType(resolvedRuntimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result)));
      }
      return completeObjectValue(exeContext, ensureValidRuntimeType(runtimeType, exeContext, returnType, fieldNodes, info, result), fieldNodes, info, path, result);
    }(exeContext, returnType, fieldNodes, info, path, result);
  }
  if (type_definition.isObjectType(returnType)) {
    return completeObjectValue(exeContext, returnType, fieldNodes, info, path, result);
  }
  jsutils_invariant.invariant(!1);
}

function ensureValidRuntimeType(runtimeTypeName, exeContext, returnType, fieldNodes, info, result) {
  if (null == runtimeTypeName) {
    throw new graphql_web.GraphQLError(`Abstract type "${returnType.name}" must resolve to an Object type at runtime for field "${info.parentType.name}.${info.fieldName}". Either the "${returnType.name}" type should provide a "resolveType" function or each possible type should provide an "isTypeOf" function.`, fieldNodes);
  }
  if (type_definition.isObjectType(runtimeTypeName)) {
    throw new graphql_web.GraphQLError("Support for returning GraphQLObjectType from resolveType was removed in graphql-js@16.0.0 please return type name instead.");
  }
  if ("string" != typeof runtimeTypeName) {
    throw new graphql_web.GraphQLError(`Abstract type "${returnType.name}" must resolve to an Object type at runtime for field "${info.parentType.name}.${info.fieldName}" with value ${jsutils_inspect.inspect(result)}, received "${jsutils_inspect.inspect(runtimeTypeName)}".`);
  }
  const runtimeType = exeContext.schema.getType(runtimeTypeName);
  if (null == runtimeType) {
    throw new graphql_web.GraphQLError(`Abstract type "${returnType.name}" was resolved to a type "${runtimeTypeName}" that does not exist inside the schema.`, {
      nodes: fieldNodes
    });
  }
  if (!type_definition.isObjectType(runtimeType)) {
    throw new graphql_web.GraphQLError(`Abstract type "${returnType.name}" was resolved to a non-object type "${runtimeTypeName}".`, {
      nodes: fieldNodes
    });
  }
  if (!exeContext.schema.isSubType(returnType, runtimeType)) {
    throw new graphql_web.GraphQLError(`Runtime Object type "${runtimeType.name}" is not a possible type for "${returnType.name}".`, {
      nodes: fieldNodes
    });
  }
  return runtimeType;
}

function completeObjectValue(exeContext, returnType, fieldNodes, info, path, result) {
  const subFieldNodes = collectSubfields(exeContext, returnType, fieldNodes);
  if (returnType.isTypeOf) {
    const isTypeOf = returnType.isTypeOf(result, exeContext.contextValue, info);
    if (jsutils_isPromise.isPromise(isTypeOf)) {
      return isTypeOf.then((resolvedIsTypeOf => {
        if (!resolvedIsTypeOf) {
          throw invalidReturnTypeError(returnType, result, fieldNodes);
        }
        return executeFields(exeContext, returnType, result, path, subFieldNodes);
      }));
    }
    if (!isTypeOf) {
      throw invalidReturnTypeError(returnType, result, fieldNodes);
    }
  }
  return executeFields(exeContext, returnType, result, path, subFieldNodes);
}

function invalidReturnTypeError(returnType, result, fieldNodes) {
  return new graphql_web.GraphQLError(`Expected value of type "${returnType.name}" but got: ${jsutils_inspect.inspect(result)}.`, {
    nodes: fieldNodes
  });
}

const defaultTypeResolver = function(value, contextValue, info, abstractType) {
  if (jsutils_isObjectLike.isObjectLike(value) && "string" == typeof value.__typename) {
    return value.__typename;
  }
  const possibleTypes = info.schema.getPossibleTypes(abstractType);
  const promisedIsTypeOfResults = [];
  for (let i = 0; i < possibleTypes.length; i++) {
    const type = possibleTypes[i];
    if (type.isTypeOf) {
      const isTypeOfResult = type.isTypeOf(value, contextValue, info);
      if (jsutils_isPromise.isPromise(isTypeOfResult)) {
        promisedIsTypeOfResults[i] = isTypeOfResult;
      } else if (isTypeOfResult) {
        return type.name;
      }
    }
  }
  if (promisedIsTypeOfResults.length) {
    return Promise.all(promisedIsTypeOfResults).then((isTypeOfResults => {
      for (let i = 0; i < isTypeOfResults.length; i++) {
        if (isTypeOfResults[i]) {
          return possibleTypes[i].name;
        }
      }
    }));
  }
};

const defaultFieldResolver = function(source, args, contextValue, info) {
  if (jsutils_isObjectLike.isObjectLike(source) || "function" == typeof source) {
    const property = source[info.fieldName];
    if ("function" == typeof property) {
      return source[info.fieldName](args, contextValue, info);
    }
    return property;
  }
};

function getFieldDef(schema, parentType, fieldNode) {
  const fieldName = fieldNode.name.value;
  if (fieldName === type_introspection.SchemaMetaFieldDef.name && schema.getQueryType() === parentType) {
    return type_introspection.SchemaMetaFieldDef;
  } else if (fieldName === type_introspection.TypeMetaFieldDef.name && schema.getQueryType() === parentType) {
    return type_introspection.TypeMetaFieldDef;
  } else if (fieldName === type_introspection.TypeNameMetaFieldDef.name) {
    return type_introspection.TypeNameMetaFieldDef;
  }
  return parentType.getFields()[fieldName];
}

exports.assertValidExecutionArguments = assertValidExecutionArguments;

exports.buildExecutionContext = buildExecutionContext;

exports.buildResolveInfo = buildResolveInfo;

exports.defaultFieldResolver = defaultFieldResolver;

exports.defaultTypeResolver = defaultTypeResolver;

exports.execute = execute;

exports.executeSync = function executeSync(args) {
  const result = execute(args);
  if (jsutils_isPromise.isPromise(result)) {
    throw new Error("GraphQL execution failed to complete synchronously.");
  }
  return result;
};

exports.getFieldDef = getFieldDef;
