import { inspect } from "../jsutils/inspect.mjs";

import { isAsyncIterable } from "../jsutils/isAsyncIterable.mjs";

import { pathToArray, addPath } from "../jsutils/Path.mjs";

import { locatedError } from "../error/locatedError.mjs";

import { collectFields } from "./collectFields.mjs";

import { assertValidExecutionArguments, buildExecutionContext, getFieldDef, buildResolveInfo, execute } from "./execute.mjs";

import { mapAsyncIterator } from "./mapAsyncIterator.mjs";

import { getArgumentValues } from "./values.mjs";

import { GraphQLError } from "@0no-co/graphql.web";

async function subscribe(args) {
  const resultOrStream = await createSourceEventStream(args);
  if (!isAsyncIterable(resultOrStream)) {
    return resultOrStream;
  }
  return mapAsyncIterator(resultOrStream, (payload => execute({
    ...args,
    rootValue: payload
  })));
}

async function createSourceEventStream(...rawArgs) {
  const args = function toNormalizedArgs(args) {
    const firstArg = args[0];
    if (firstArg && "document" in firstArg) {
      return firstArg;
    }
    return {
      schema: firstArg,
      document: args[1],
      rootValue: args[2],
      contextValue: args[3],
      variableValues: args[4],
      operationName: args[5],
      subscribeFieldResolver: args[6]
    };
  }(rawArgs);
  const {schema} = args;
  assertValidExecutionArguments(schema);
  const exeContext = buildExecutionContext(args);
  if (!("schema" in exeContext)) {
    return {
      errors: exeContext
    };
  }
  try {
    const eventStream = await async function executeSubscription(exeContext) {
      const {schema, fragments, operation, variableValues, rootValue} = exeContext;
      const rootType = schema.getSubscriptionType();
      if (null == rootType) {
        throw new GraphQLError("Schema is not configured to execute subscription operation.", {
          nodes: operation
        });
      }
      const rootFields = collectFields(schema, fragments, variableValues, rootType, operation.selectionSet);
      const [responseName, fieldNodes] = [ ...rootFields.entries() ][0];
      const fieldDef = getFieldDef(schema, rootType, fieldNodes[0]);
      if (!fieldDef) {
        throw new GraphQLError(`The subscription field "${fieldNodes[0].name.value}" is not defined.`, {
          nodes: fieldNodes
        });
      }
      const path = addPath(void 0, responseName, rootType.name);
      const info = buildResolveInfo(exeContext, fieldDef, fieldNodes, rootType, path);
      try {
        var _fieldDef$subscribe;
        const args = getArgumentValues(fieldDef, fieldNodes[0], variableValues);
        const contextValue = exeContext.contextValue;
        const resolveFn = null !== (_fieldDef$subscribe = fieldDef.subscribe) && void 0 !== _fieldDef$subscribe ? _fieldDef$subscribe : exeContext.subscribeFieldResolver;
        const eventStream = await resolveFn(rootValue, args, contextValue, info);
        if (eventStream instanceof Error) {
          throw eventStream;
        }
        return eventStream;
      } catch (error) {
        throw locatedError(error, fieldNodes, pathToArray(path));
      }
    }(exeContext);
    if (!isAsyncIterable(eventStream)) {
      throw new Error(`Subscription field must return Async Iterable. Received: ${inspect(eventStream)}.`);
    }
    return eventStream;
  } catch (error) {
    if (error instanceof GraphQLError) {
      return {
        errors: [ error ]
      };
    }
    throw error;
  }
}

export { createSourceEventStream, subscribe };
