import { inspect } from "../jsutils/inspect.mjs";

import { keyMap } from "../jsutils/keyMap.mjs";

import { printPathArray } from "../jsutils/printPathArray.mjs";

import { isInputType, isNonNullType } from "../type/definition.mjs";

import { coerceInputValue } from "../utilities/coerceInputValue.mjs";

import { typeFromAST } from "../utilities/typeFromAST.mjs";

import { valueFromAST } from "../utilities/valueFromAST.mjs";

import { print, GraphQLError, Kind } from "@0no-co/graphql.web";

function getVariableValues(schema, varDefNodes, inputs, options) {
  const errors = [];
  const maxErrors = null == options ? void 0 : options.maxErrors;
  try {
    const coerced = function coerceVariableValues(schema, varDefNodes, inputs, onError) {
      const coercedValues = {};
      for (const varDefNode of varDefNodes) {
        const varName = varDefNode.variable.name.value;
        const varType = typeFromAST(schema, varDefNode.type);
        if (!isInputType(varType)) {
          const varTypeStr = print(varDefNode.type);
          onError(new GraphQLError(`Variable "$${varName}" expected value of type "${varTypeStr}" which cannot be used as an input type.`, {
            nodes: varDefNode.type
          }));
          continue;
        }
        if (!hasOwnProperty(inputs, varName)) {
          if (varDefNode.defaultValue) {
            coercedValues[varName] = valueFromAST(varDefNode.defaultValue, varType);
          } else if (isNonNullType(varType)) {
            const varTypeStr = inspect(varType);
            onError(new GraphQLError(`Variable "$${varName}" of required type "${varTypeStr}" was not provided.`, {
              nodes: varDefNode
            }));
          }
          continue;
        }
        const value = inputs[varName];
        if (null === value && isNonNullType(varType)) {
          const varTypeStr = inspect(varType);
          onError(new GraphQLError(`Variable "$${varName}" of non-null type "${varTypeStr}" must not be null.`, {
            nodes: varDefNode
          }));
          continue;
        }
        coercedValues[varName] = coerceInputValue(value, varType, ((path, invalidValue, error) => {
          let prefix = `Variable "$${varName}" got invalid value ` + inspect(invalidValue);
          if (path.length > 0) {
            prefix += ` at "${varName}${printPathArray(path)}"`;
          }
          onError(new GraphQLError(prefix + "; " + error.message, {
            nodes: varDefNode,
            originalError: error.originalError
          }));
        }));
      }
      return coercedValues;
    }(schema, varDefNodes, inputs, (error => {
      if (null != maxErrors && errors.length >= maxErrors) {
        throw new GraphQLError("Too many errors processing variables, error limit reached. Execution aborted.");
      }
      errors.push(error);
    }));
    if (0 === errors.length) {
      return {
        coerced
      };
    }
  } catch (error) {
    errors.push(error);
  }
  return {
    errors
  };
}

function getArgumentValues(def, node, variableValues) {
  var _node$arguments;
  const coercedValues = {};
  const argumentNodes = null !== (_node$arguments = node.arguments) && void 0 !== _node$arguments ? _node$arguments : [];
  const argNodeMap = keyMap(argumentNodes, (arg => arg.name.value));
  for (const argDef of def.args) {
    const name = argDef.name;
    const argType = argDef.type;
    const argumentNode = argNodeMap[name];
    if (!argumentNode) {
      if (void 0 !== argDef.defaultValue) {
        coercedValues[name] = argDef.defaultValue;
      } else if (isNonNullType(argType)) {
        throw new GraphQLError(`Argument "${name}" of required type "${inspect(argType)}" was not provided.`, {
          nodes: node
        });
      }
      continue;
    }
    const valueNode = argumentNode.value;
    let isNull = valueNode.kind === Kind.NULL;
    if (valueNode.kind === Kind.VARIABLE) {
      const variableName = valueNode.name.value;
      if (null == variableValues || !hasOwnProperty(variableValues, variableName)) {
        if (void 0 !== argDef.defaultValue) {
          coercedValues[name] = argDef.defaultValue;
        } else if (isNonNullType(argType)) {
          throw new GraphQLError(`Argument "${name}" of required type "${inspect(argType)}" was provided the variable "$${variableName}" which was not provided a runtime value.`, {
            nodes: valueNode
          });
        }
        continue;
      }
      isNull = null == variableValues[variableName];
    }
    if (isNull && isNonNullType(argType)) {
      throw new GraphQLError(`Argument "${name}" of non-null type "${inspect(argType)}" must not be null.`, {
        nodes: valueNode
      });
    }
    const coercedValue = valueFromAST(valueNode, argType, variableValues);
    if (void 0 === coercedValue) {
      throw new GraphQLError(`Argument "${name}" has invalid value ${print(valueNode)}.`, {
        nodes: valueNode
      });
    }
    coercedValues[name] = coercedValue;
  }
  return coercedValues;
}

function getDirectiveValues(directiveDef, node, variableValues) {
  var _node$directives;
  const directiveNode = null === (_node$directives = node.directives) || void 0 === _node$directives ? void 0 : _node$directives.find((directive => directive.name.value === directiveDef.name));
  if (directiveNode) {
    return getArgumentValues(directiveDef, directiveNode, variableValues);
  }
}

function hasOwnProperty(obj, prop) {
  return Object.prototype.hasOwnProperty.call(obj, prop);
}

export { getArgumentValues, getDirectiveValues, getVariableValues };
