var language_ast = require("../language/ast.js");

var language_visitor = require("../language/visitor.js");

var type_definition = require("../type/definition.js");

var type_introspection = require("../type/introspection.js");

var utilities_typeFromAST = require("./typeFromAST.js");

var graphql_web = require("@0no-co/graphql.web");

function getFieldDef(schema, parentType, fieldNode) {
  const name = fieldNode.name.value;
  if (name === type_introspection.SchemaMetaFieldDef.name && schema.getQueryType() === parentType) {
    return type_introspection.SchemaMetaFieldDef;
  }
  if (name === type_introspection.TypeMetaFieldDef.name && schema.getQueryType() === parentType) {
    return type_introspection.TypeMetaFieldDef;
  }
  if (name === type_introspection.TypeNameMetaFieldDef.name && type_definition.isCompositeType(parentType)) {
    return type_introspection.TypeNameMetaFieldDef;
  }
  if (type_definition.isObjectType(parentType) || type_definition.isInterfaceType(parentType)) {
    return parentType.getFields()[name];
  }
}

exports.TypeInfo = class TypeInfo {
  constructor(schema, initialType, getFieldDefFn) {
    this._schema = schema;
    this._typeStack = [];
    this._parentTypeStack = [];
    this._inputTypeStack = [];
    this._fieldDefStack = [];
    this._defaultValueStack = [];
    this._directive = null;
    this._argument = null;
    this._enumValue = null;
    this._getFieldDef = null != getFieldDefFn ? getFieldDefFn : getFieldDef;
    if (initialType) {
      if (type_definition.isInputType(initialType)) {
        this._inputTypeStack.push(initialType);
      }
      if (type_definition.isCompositeType(initialType)) {
        this._parentTypeStack.push(initialType);
      }
      if (type_definition.isOutputType(initialType)) {
        this._typeStack.push(initialType);
      }
    }
  }
  getType() {
    if (this._typeStack.length > 0) {
      return this._typeStack[this._typeStack.length - 1];
    }
  }
  getParentType() {
    if (this._parentTypeStack.length > 0) {
      return this._parentTypeStack[this._parentTypeStack.length - 1];
    }
  }
  getInputType() {
    if (this._inputTypeStack.length > 0) {
      return this._inputTypeStack[this._inputTypeStack.length - 1];
    }
  }
  getParentInputType() {
    if (this._inputTypeStack.length > 1) {
      return this._inputTypeStack[this._inputTypeStack.length - 2];
    }
  }
  getFieldDef() {
    if (this._fieldDefStack.length > 0) {
      return this._fieldDefStack[this._fieldDefStack.length - 1];
    }
  }
  getDefaultValue() {
    if (this._defaultValueStack.length > 0) {
      return this._defaultValueStack[this._defaultValueStack.length - 1];
    }
  }
  getDirective() {
    return this._directive;
  }
  getArgument() {
    return this._argument;
  }
  getEnumValue() {
    return this._enumValue;
  }
  enter(node) {
    const schema = this._schema;
    switch (node.kind) {
     case graphql_web.Kind.SELECTION_SET:
      {
        const namedType = type_definition.getNamedType(this.getType());
        this._parentTypeStack.push(type_definition.isCompositeType(namedType) ? namedType : void 0);
        break;
      }

     case graphql_web.Kind.FIELD:
      {
        const parentType = this.getParentType();
        let fieldDef;
        let fieldType;
        if (parentType) {
          fieldDef = this._getFieldDef(schema, parentType, node);
          if (fieldDef) {
            fieldType = fieldDef.type;
          }
        }
        this._fieldDefStack.push(fieldDef);
        this._typeStack.push(type_definition.isOutputType(fieldType) ? fieldType : void 0);
        break;
      }

     case graphql_web.Kind.DIRECTIVE:
      this._directive = schema.getDirective(node.name.value);
      break;

     case graphql_web.Kind.OPERATION_DEFINITION:
      {
        const rootType = schema.getRootType(node.operation);
        this._typeStack.push(type_definition.isObjectType(rootType) ? rootType : void 0);
        break;
      }

     case graphql_web.Kind.INLINE_FRAGMENT:
     case graphql_web.Kind.FRAGMENT_DEFINITION:
      {
        const typeConditionAST = node.typeCondition;
        const outputType = typeConditionAST ? utilities_typeFromAST.typeFromAST(schema, typeConditionAST) : type_definition.getNamedType(this.getType());
        this._typeStack.push(type_definition.isOutputType(outputType) ? outputType : void 0);
        break;
      }

     case graphql_web.Kind.VARIABLE_DEFINITION:
      {
        const inputType = utilities_typeFromAST.typeFromAST(schema, node.type);
        this._inputTypeStack.push(type_definition.isInputType(inputType) ? inputType : void 0);
        break;
      }

     case graphql_web.Kind.ARGUMENT:
      {
        var _this$getDirective;
        let argDef;
        let argType;
        const fieldOrDirective = null !== (_this$getDirective = this.getDirective()) && void 0 !== _this$getDirective ? _this$getDirective : this.getFieldDef();
        if (fieldOrDirective) {
          argDef = fieldOrDirective.args.find((arg => arg.name === node.name.value));
          if (argDef) {
            argType = argDef.type;
          }
        }
        this._argument = argDef;
        this._defaultValueStack.push(argDef ? argDef.defaultValue : void 0);
        this._inputTypeStack.push(type_definition.isInputType(argType) ? argType : void 0);
        break;
      }

     case graphql_web.Kind.LIST:
      {
        const listType = type_definition.getNullableType(this.getInputType());
        const itemType = type_definition.isListType(listType) ? listType.ofType : listType;
        this._defaultValueStack.push(void 0);
        this._inputTypeStack.push(type_definition.isInputType(itemType) ? itemType : void 0);
        break;
      }

     case graphql_web.Kind.OBJECT_FIELD:
      {
        const objectType = type_definition.getNamedType(this.getInputType());
        let inputFieldType;
        let inputField;
        if (type_definition.isInputObjectType(objectType)) {
          inputField = objectType.getFields()[node.name.value];
          if (inputField) {
            inputFieldType = inputField.type;
          }
        }
        this._defaultValueStack.push(inputField ? inputField.defaultValue : void 0);
        this._inputTypeStack.push(type_definition.isInputType(inputFieldType) ? inputFieldType : void 0);
        break;
      }

     case graphql_web.Kind.ENUM:
      {
        const enumType = type_definition.getNamedType(this.getInputType());
        let enumValue;
        if (type_definition.isEnumType(enumType)) {
          enumValue = enumType.getValue(node.value);
        }
        this._enumValue = enumValue;
        break;
      }
    }
  }
  leave(node) {
    switch (node.kind) {
     case graphql_web.Kind.SELECTION_SET:
      this._parentTypeStack.pop();
      break;

     case graphql_web.Kind.FIELD:
      this._fieldDefStack.pop();
      this._typeStack.pop();
      break;

     case graphql_web.Kind.DIRECTIVE:
      this._directive = null;
      break;

     case graphql_web.Kind.OPERATION_DEFINITION:
     case graphql_web.Kind.INLINE_FRAGMENT:
     case graphql_web.Kind.FRAGMENT_DEFINITION:
      this._typeStack.pop();
      break;

     case graphql_web.Kind.VARIABLE_DEFINITION:
      this._inputTypeStack.pop();
      break;

     case graphql_web.Kind.ARGUMENT:
      this._argument = null;
      this._defaultValueStack.pop();
      this._inputTypeStack.pop();
      break;

     case graphql_web.Kind.LIST:
     case graphql_web.Kind.OBJECT_FIELD:
      this._defaultValueStack.pop();
      this._inputTypeStack.pop();
      break;

     case graphql_web.Kind.ENUM:
      this._enumValue = null;
    }
  }
};

exports.visitWithTypeInfo = function visitWithTypeInfo(typeInfo, visitor) {
  return {
    enter(...args) {
      const node = args[0];
      typeInfo.enter(node);
      const fn = language_visitor.getEnterLeaveForKind(visitor, node.kind).enter;
      if (fn) {
        const result = fn.apply(visitor, args);
        if (void 0 !== result) {
          typeInfo.leave(node);
          if (language_ast.isNode(result)) {
            typeInfo.enter(result);
          }
        }
        return result;
      }
    },
    leave(...args) {
      const node = args[0];
      const fn = language_visitor.getEnterLeaveForKind(visitor, node.kind).leave;
      let result;
      if (fn) {
        result = fn.apply(visitor, args);
      }
      typeInfo.leave(node);
      return result;
    }
  };
};
