import { invariant } from "../jsutils/invariant.mjs";

import { keyMap } from "../jsutils/keyMap.mjs";

import { mapValue } from "../jsutils/mapValue.mjs";

import { isTypeDefinitionNode, isTypeExtensionNode } from "../language/predicates.mjs";

import { isScalarType, isObjectType, isInterfaceType, isUnionType, isEnumType, isInputObjectType, GraphQLInputObjectType, GraphQLEnumType, GraphQLScalarType, GraphQLObjectType, GraphQLInterfaceType, GraphQLUnionType, GraphQLList, GraphQLNonNull, isListType, isNonNullType } from "../type/definition.mjs";

import { GraphQLDirective, GraphQLDeprecatedDirective, GraphQLSpecifiedByDirective } from "../type/directives.mjs";

import { isIntrospectionType, introspectionTypes } from "../type/introspection.mjs";

import { isSpecifiedScalarType, specifiedScalarTypes } from "../type/scalars.mjs";

import { assertSchema, GraphQLSchema } from "../type/schema.mjs";

import { assertValidSDLExtension } from "../validation/validate.mjs";

import { getDirectiveValues } from "../execution/values.mjs";

import { valueFromAST } from "./valueFromAST.mjs";

import { Kind } from "@0no-co/graphql.web";

function extendSchema(schema, documentAST, options) {
  assertSchema(schema);
  if (!0 !== (null == options ? void 0 : options.assumeValid) && !0 !== (null == options ? void 0 : options.assumeValidSDL)) {
    assertValidSDLExtension(documentAST, schema);
  }
  const schemaConfig = schema.toConfig();
  const extendedConfig = extendSchemaImpl(schemaConfig, documentAST, options);
  return schemaConfig === extendedConfig ? schema : new GraphQLSchema(extendedConfig);
}

function extendSchemaImpl(schemaConfig, documentAST, options) {
  var _schemaDef, _schemaDef$descriptio, _schemaDef2, _options$assumeValid;
  const typeDefs = [];
  const typeExtensionsMap = Object.create(null);
  const directiveDefs = [];
  let schemaDef;
  const schemaExtensions = [];
  for (const def of documentAST.definitions) {
    if (def.kind === Kind.SCHEMA_DEFINITION) {
      schemaDef = def;
    } else if (def.kind === Kind.SCHEMA_EXTENSION) {
      schemaExtensions.push(def);
    } else if (isTypeDefinitionNode(def)) {
      typeDefs.push(def);
    } else if (isTypeExtensionNode(def)) {
      const extendedTypeName = def.name.value;
      const existingTypeExtensions = typeExtensionsMap[extendedTypeName];
      typeExtensionsMap[extendedTypeName] = existingTypeExtensions ? existingTypeExtensions.concat([ def ]) : [ def ];
    } else if (def.kind === Kind.DIRECTIVE_DEFINITION) {
      directiveDefs.push(def);
    }
  }
  if (0 === Object.keys(typeExtensionsMap).length && 0 === typeDefs.length && 0 === directiveDefs.length && 0 === schemaExtensions.length && null == schemaDef) {
    return schemaConfig;
  }
  const typeMap = Object.create(null);
  for (const existingType of schemaConfig.types) {
    typeMap[existingType.name] = extendNamedType(existingType);
  }
  for (const typeNode of typeDefs) {
    var _stdTypeMap$name;
    const name = typeNode.name.value;
    typeMap[name] = null !== (_stdTypeMap$name = stdTypeMap[name]) && void 0 !== _stdTypeMap$name ? _stdTypeMap$name : buildType(typeNode);
  }
  const operationTypes = {
    query: schemaConfig.query && replaceNamedType(schemaConfig.query),
    mutation: schemaConfig.mutation && replaceNamedType(schemaConfig.mutation),
    subscription: schemaConfig.subscription && replaceNamedType(schemaConfig.subscription),
    ...schemaDef && getOperationTypes([ schemaDef ]),
    ...getOperationTypes(schemaExtensions)
  };
  return {
    description: null === (_schemaDef = schemaDef) || void 0 === _schemaDef ? void 0 : null === (_schemaDef$descriptio = _schemaDef.description) || void 0 === _schemaDef$descriptio ? void 0 : _schemaDef$descriptio.value,
    ...operationTypes,
    types: Object.values(typeMap),
    directives: [ ...schemaConfig.directives.map((function replaceDirective(directive) {
      const config = directive.toConfig();
      return new GraphQLDirective({
        ...config,
        args: mapValue(config.args, extendArg)
      });
    })), ...directiveDefs.map((function buildDirective(node) {
      var _node$description;
      return new GraphQLDirective({
        name: node.name.value,
        description: null === (_node$description = node.description) || void 0 === _node$description ? void 0 : _node$description.value,
        locations: node.locations.map((({value}) => value)),
        isRepeatable: node.repeatable,
        args: buildArgumentMap(node.arguments),
        astNode: node
      });
    })) ],
    extensions: Object.create(null),
    astNode: null !== (_schemaDef2 = schemaDef) && void 0 !== _schemaDef2 ? _schemaDef2 : schemaConfig.astNode,
    extensionASTNodes: schemaConfig.extensionASTNodes.concat(schemaExtensions),
    assumeValid: null !== (_options$assumeValid = null == options ? void 0 : options.assumeValid) && void 0 !== _options$assumeValid ? _options$assumeValid : !1
  };
  function replaceType(type) {
    if (isListType(type)) {
      return new GraphQLList(replaceType(type.ofType));
    }
    if (isNonNullType(type)) {
      return new GraphQLNonNull(replaceType(type.ofType));
    }
    return replaceNamedType(type);
  }
  function replaceNamedType(type) {
    return typeMap[type.name];
  }
  function extendNamedType(type) {
    if (isIntrospectionType(type) || isSpecifiedScalarType(type)) {
      return type;
    }
    if (isScalarType(type)) {
      return function extendScalarType(type) {
        var _typeExtensionsMap$co2;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$co2 = typeExtensionsMap[config.name]) && void 0 !== _typeExtensionsMap$co2 ? _typeExtensionsMap$co2 : [];
        let specifiedByURL = config.specifiedByURL;
        for (const extensionNode of extensions) {
          var _getSpecifiedByURL;
          specifiedByURL = null !== (_getSpecifiedByURL = getSpecifiedByURL(extensionNode)) && void 0 !== _getSpecifiedByURL ? _getSpecifiedByURL : specifiedByURL;
        }
        return new GraphQLScalarType({
          ...config,
          specifiedByURL,
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    if (isObjectType(type)) {
      return function extendObjectType(type) {
        var _typeExtensionsMap$co3;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$co3 = typeExtensionsMap[config.name]) && void 0 !== _typeExtensionsMap$co3 ? _typeExtensionsMap$co3 : [];
        return new GraphQLObjectType({
          ...config,
          interfaces: () => [ ...type.getInterfaces().map(replaceNamedType), ...buildInterfaces(extensions) ],
          fields: () => ({
            ...mapValue(config.fields, extendField),
            ...buildFieldMap(extensions)
          }),
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    if (isInterfaceType(type)) {
      return function extendInterfaceType(type) {
        var _typeExtensionsMap$co4;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$co4 = typeExtensionsMap[config.name]) && void 0 !== _typeExtensionsMap$co4 ? _typeExtensionsMap$co4 : [];
        return new GraphQLInterfaceType({
          ...config,
          interfaces: () => [ ...type.getInterfaces().map(replaceNamedType), ...buildInterfaces(extensions) ],
          fields: () => ({
            ...mapValue(config.fields, extendField),
            ...buildFieldMap(extensions)
          }),
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    if (isUnionType(type)) {
      return function extendUnionType(type) {
        var _typeExtensionsMap$co5;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$co5 = typeExtensionsMap[config.name]) && void 0 !== _typeExtensionsMap$co5 ? _typeExtensionsMap$co5 : [];
        return new GraphQLUnionType({
          ...config,
          types: () => [ ...type.getTypes().map(replaceNamedType), ...buildUnionTypes(extensions) ],
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    if (isEnumType(type)) {
      return function extendEnumType(type) {
        var _typeExtensionsMap$ty;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$ty = typeExtensionsMap[type.name]) && void 0 !== _typeExtensionsMap$ty ? _typeExtensionsMap$ty : [];
        return new GraphQLEnumType({
          ...config,
          values: {
            ...config.values,
            ...buildEnumValueMap(extensions)
          },
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    if (isInputObjectType(type)) {
      return function extendInputObjectType(type) {
        var _typeExtensionsMap$co;
        const config = type.toConfig();
        const extensions = null !== (_typeExtensionsMap$co = typeExtensionsMap[config.name]) && void 0 !== _typeExtensionsMap$co ? _typeExtensionsMap$co : [];
        return new GraphQLInputObjectType({
          ...config,
          fields: () => ({
            ...mapValue(config.fields, (field => ({
              ...field,
              type: replaceType(field.type)
            }))),
            ...buildInputFieldMap(extensions)
          }),
          extensionASTNodes: config.extensionASTNodes.concat(extensions)
        });
      }(type);
    }
    invariant(!1);
  }
  function extendField(field) {
    return {
      ...field,
      type: replaceType(field.type),
      args: field.args && mapValue(field.args, extendArg)
    };
  }
  function extendArg(arg) {
    return {
      ...arg,
      type: replaceType(arg.type)
    };
  }
  function getOperationTypes(nodes) {
    const opTypes = {};
    for (const node of nodes) {
      var _node$operationTypes;
      const operationTypesNodes = null !== (_node$operationTypes = node.operationTypes) && void 0 !== _node$operationTypes ? _node$operationTypes : [];
      for (const operationType of operationTypesNodes) {
        opTypes[operationType.operation] = getNamedType(operationType.type);
      }
    }
    return opTypes;
  }
  function getNamedType(node) {
    var _stdTypeMap$name2;
    const name = node.name.value;
    const type = null !== (_stdTypeMap$name2 = stdTypeMap[name]) && void 0 !== _stdTypeMap$name2 ? _stdTypeMap$name2 : typeMap[name];
    if (void 0 === type) {
      throw new Error(`Unknown type: "${name}".`);
    }
    return type;
  }
  function getWrappedType(node) {
    if (node.kind === Kind.LIST_TYPE) {
      return new GraphQLList(getWrappedType(node.type));
    }
    if (node.kind === Kind.NON_NULL_TYPE) {
      return new GraphQLNonNull(getWrappedType(node.type));
    }
    return getNamedType(node);
  }
  function buildFieldMap(nodes) {
    const fieldConfigMap = Object.create(null);
    for (const node of nodes) {
      var _node$fields;
      const nodeFields = null !== (_node$fields = node.fields) && void 0 !== _node$fields ? _node$fields : [];
      for (const field of nodeFields) {
        var _field$description;
        fieldConfigMap[field.name.value] = {
          type: getWrappedType(field.type),
          description: null === (_field$description = field.description) || void 0 === _field$description ? void 0 : _field$description.value,
          args: buildArgumentMap(field.arguments),
          deprecationReason: getDeprecationReason(field),
          astNode: field
        };
      }
    }
    return fieldConfigMap;
  }
  function buildArgumentMap(args) {
    const argsNodes = null != args ? args : [];
    const argConfigMap = Object.create(null);
    for (const arg of argsNodes) {
      var _arg$description;
      const type = getWrappedType(arg.type);
      argConfigMap[arg.name.value] = {
        type,
        description: null === (_arg$description = arg.description) || void 0 === _arg$description ? void 0 : _arg$description.value,
        defaultValue: valueFromAST(arg.defaultValue, type),
        deprecationReason: getDeprecationReason(arg),
        astNode: arg
      };
    }
    return argConfigMap;
  }
  function buildInputFieldMap(nodes) {
    const inputFieldMap = Object.create(null);
    for (const node of nodes) {
      var _node$fields2;
      const fieldsNodes = null !== (_node$fields2 = node.fields) && void 0 !== _node$fields2 ? _node$fields2 : [];
      for (const field of fieldsNodes) {
        var _field$description2;
        const type = getWrappedType(field.type);
        inputFieldMap[field.name.value] = {
          type,
          description: null === (_field$description2 = field.description) || void 0 === _field$description2 ? void 0 : _field$description2.value,
          defaultValue: valueFromAST(field.defaultValue, type),
          deprecationReason: getDeprecationReason(field),
          astNode: field
        };
      }
    }
    return inputFieldMap;
  }
  function buildEnumValueMap(nodes) {
    const enumValueMap = Object.create(null);
    for (const node of nodes) {
      var _node$values;
      const valuesNodes = null !== (_node$values = node.values) && void 0 !== _node$values ? _node$values : [];
      for (const value of valuesNodes) {
        var _value$description;
        enumValueMap[value.name.value] = {
          description: null === (_value$description = value.description) || void 0 === _value$description ? void 0 : _value$description.value,
          deprecationReason: getDeprecationReason(value),
          astNode: value
        };
      }
    }
    return enumValueMap;
  }
  function buildInterfaces(nodes) {
    return nodes.flatMap((node => {
      var _node$interfaces$map, _node$interfaces;
      return null !== (_node$interfaces$map = null === (_node$interfaces = node.interfaces) || void 0 === _node$interfaces ? void 0 : _node$interfaces.map(getNamedType)) && void 0 !== _node$interfaces$map ? _node$interfaces$map : [];
    }));
  }
  function buildUnionTypes(nodes) {
    return nodes.flatMap((node => {
      var _node$types$map, _node$types;
      return null !== (_node$types$map = null === (_node$types = node.types) || void 0 === _node$types ? void 0 : _node$types.map(getNamedType)) && void 0 !== _node$types$map ? _node$types$map : [];
    }));
  }
  function buildType(astNode) {
    var _typeExtensionsMap$na;
    const name = astNode.name.value;
    const extensionASTNodes = null !== (_typeExtensionsMap$na = typeExtensionsMap[name]) && void 0 !== _typeExtensionsMap$na ? _typeExtensionsMap$na : [];
    switch (astNode.kind) {
     case Kind.OBJECT_TYPE_DEFINITION:
      {
        var _astNode$description;
        const allNodes = [ astNode, ...extensionASTNodes ];
        return new GraphQLObjectType({
          name,
          description: null === (_astNode$description = astNode.description) || void 0 === _astNode$description ? void 0 : _astNode$description.value,
          interfaces: () => buildInterfaces(allNodes),
          fields: () => buildFieldMap(allNodes),
          astNode,
          extensionASTNodes
        });
      }

     case Kind.INTERFACE_TYPE_DEFINITION:
      {
        var _astNode$description2;
        const allNodes = [ astNode, ...extensionASTNodes ];
        return new GraphQLInterfaceType({
          name,
          description: null === (_astNode$description2 = astNode.description) || void 0 === _astNode$description2 ? void 0 : _astNode$description2.value,
          interfaces: () => buildInterfaces(allNodes),
          fields: () => buildFieldMap(allNodes),
          astNode,
          extensionASTNodes
        });
      }

     case Kind.ENUM_TYPE_DEFINITION:
      {
        var _astNode$description3;
        const allNodes = [ astNode, ...extensionASTNodes ];
        return new GraphQLEnumType({
          name,
          description: null === (_astNode$description3 = astNode.description) || void 0 === _astNode$description3 ? void 0 : _astNode$description3.value,
          values: buildEnumValueMap(allNodes),
          astNode,
          extensionASTNodes
        });
      }

     case Kind.UNION_TYPE_DEFINITION:
      {
        var _astNode$description4;
        const allNodes = [ astNode, ...extensionASTNodes ];
        return new GraphQLUnionType({
          name,
          description: null === (_astNode$description4 = astNode.description) || void 0 === _astNode$description4 ? void 0 : _astNode$description4.value,
          types: () => buildUnionTypes(allNodes),
          astNode,
          extensionASTNodes
        });
      }

     case Kind.SCALAR_TYPE_DEFINITION:
      var _astNode$description5;
      return new GraphQLScalarType({
        name,
        description: null === (_astNode$description5 = astNode.description) || void 0 === _astNode$description5 ? void 0 : _astNode$description5.value,
        specifiedByURL: getSpecifiedByURL(astNode),
        astNode,
        extensionASTNodes
      });

     case Kind.INPUT_OBJECT_TYPE_DEFINITION:
      {
        var _astNode$description6;
        const allNodes = [ astNode, ...extensionASTNodes ];
        return new GraphQLInputObjectType({
          name,
          description: null === (_astNode$description6 = astNode.description) || void 0 === _astNode$description6 ? void 0 : _astNode$description6.value,
          fields: () => buildInputFieldMap(allNodes),
          astNode,
          extensionASTNodes
        });
      }
    }
  }
}

const stdTypeMap = keyMap([ ...specifiedScalarTypes, ...introspectionTypes ], (type => type.name));

function getDeprecationReason(node) {
  const deprecated = getDirectiveValues(GraphQLDeprecatedDirective, node);
  return null == deprecated ? void 0 : deprecated.reason;
}

function getSpecifiedByURL(node) {
  const specifiedBy = getDirectiveValues(GraphQLSpecifiedByDirective, node);
  return null == specifiedBy ? void 0 : specifiedBy.url;
}

export { extendSchema, extendSchemaImpl };
