import { invariant } from "../jsutils/invariant.mjs";

import { isPrintableAsBlockString } from "../language/blockString.mjs";

import { isScalarType, isObjectType, isInterfaceType, isUnionType, isEnumType, isInputObjectType } from "../type/definition.mjs";

import { DEFAULT_DEPRECATION_REASON, isSpecifiedDirective } from "../type/directives.mjs";

import { isIntrospectionType } from "../type/introspection.mjs";

import { isSpecifiedScalarType } from "../type/scalars.mjs";

import { astFromValue } from "./astFromValue.mjs";

import { print, Kind } from "@0no-co/graphql.web";

function printSchema(schema) {
  return printFilteredSchema(schema, (n => !isSpecifiedDirective(n)), isDefinedType);
}

function printIntrospectionSchema(schema) {
  return printFilteredSchema(schema, isSpecifiedDirective, isIntrospectionType);
}

function isDefinedType(type) {
  return !isSpecifiedScalarType(type) && !isIntrospectionType(type);
}

function printFilteredSchema(schema, directiveFilter, typeFilter) {
  const directives = schema.getDirectives().filter(directiveFilter);
  const types = Object.values(schema.getTypeMap()).filter(typeFilter);
  return [ printSchemaDefinition(schema), ...directives.map((directive => function printDirective(directive) {
    return printDescription(directive) + "directive @" + directive.name + printArgs(directive.args) + (directive.isRepeatable ? " repeatable" : "") + " on " + directive.locations.join(" | ");
  }(directive))), ...types.map((type => printType(type))) ].filter(Boolean).join("\n\n");
}

function printSchemaDefinition(schema) {
  if (null == schema.description && function isSchemaOfCommonNames(schema) {
    const queryType = schema.getQueryType();
    if (queryType && "Query" !== queryType.name) {
      return !1;
    }
    const mutationType = schema.getMutationType();
    if (mutationType && "Mutation" !== mutationType.name) {
      return !1;
    }
    const subscriptionType = schema.getSubscriptionType();
    if (subscriptionType && "Subscription" !== subscriptionType.name) {
      return !1;
    }
    return !0;
  }(schema)) {
    return;
  }
  const operationTypes = [];
  const queryType = schema.getQueryType();
  if (queryType) {
    operationTypes.push(`  query: ${queryType.name}`);
  }
  const mutationType = schema.getMutationType();
  if (mutationType) {
    operationTypes.push(`  mutation: ${mutationType.name}`);
  }
  const subscriptionType = schema.getSubscriptionType();
  if (subscriptionType) {
    operationTypes.push(`  subscription: ${subscriptionType.name}`);
  }
  return printDescription(schema) + `schema {\n${operationTypes.join("\n")}\n}`;
}

function printType(type) {
  if (isScalarType(type)) {
    return function printScalar(type) {
      return printDescription(type) + `scalar ${type.name}` + function printSpecifiedByURL(scalar) {
        if (null == scalar.specifiedByURL) {
          return "";
        }
        return ` @specifiedBy(url: ${print({
          kind: Kind.STRING,
          value: scalar.specifiedByURL
        })})`;
      }(type);
    }(type);
  }
  if (isObjectType(type)) {
    return function printObject(type) {
      return printDescription(type) + `type ${type.name}` + printImplementedInterfaces(type) + printFields(type);
    }(type);
  }
  if (isInterfaceType(type)) {
    return function printInterface(type) {
      return printDescription(type) + `interface ${type.name}` + printImplementedInterfaces(type) + printFields(type);
    }(type);
  }
  if (isUnionType(type)) {
    return function printUnion(type) {
      const types = type.getTypes();
      const possibleTypes = types.length ? " = " + types.join(" | ") : "";
      return printDescription(type) + "union " + type.name + possibleTypes;
    }(type);
  }
  if (isEnumType(type)) {
    return function printEnum(type) {
      const values = type.getValues().map(((value, i) => printDescription(value, "  ", !i) + "  " + value.name + printDeprecated(value.deprecationReason)));
      return printDescription(type) + `enum ${type.name}` + printBlock(values);
    }(type);
  }
  if (isInputObjectType(type)) {
    return function printInputObject(type) {
      const fields = Object.values(type.getFields()).map(((f, i) => printDescription(f, "  ", !i) + "  " + printInputValue(f)));
      return printDescription(type) + `input ${type.name}` + printBlock(fields);
    }(type);
  }
  invariant(!1);
}

function printImplementedInterfaces(type) {
  const interfaces = type.getInterfaces();
  return interfaces.length ? " implements " + interfaces.map((i => i.name)).join(" & ") : "";
}

function printFields(type) {
  return printBlock(Object.values(type.getFields()).map(((f, i) => printDescription(f, "  ", !i) + "  " + f.name + printArgs(f.args, "  ") + ": " + String(f.type) + printDeprecated(f.deprecationReason))));
}

function printBlock(items) {
  return 0 !== items.length ? " {\n" + items.join("\n") + "\n}" : "";
}

function printArgs(args, indentation = "") {
  if (0 === args.length) {
    return "";
  }
  if (args.every((arg => !arg.description))) {
    return "(" + args.map(printInputValue).join(", ") + ")";
  }
  return "(\n" + args.map(((arg, i) => printDescription(arg, "  " + indentation, !i) + "  " + indentation + printInputValue(arg))).join("\n") + "\n" + indentation + ")";
}

function printInputValue(arg) {
  const defaultAST = astFromValue(arg.defaultValue, arg.type);
  let argDecl = arg.name + ": " + String(arg.type);
  if (defaultAST) {
    argDecl += ` = ${print(defaultAST)}`;
  }
  return argDecl + printDeprecated(arg.deprecationReason);
}

function printDeprecated(reason) {
  if (null == reason) {
    return "";
  }
  if (reason !== DEFAULT_DEPRECATION_REASON) {
    return ` @deprecated(reason: ${print({
      kind: Kind.STRING,
      value: reason
    })})`;
  }
  return " @deprecated";
}

function printDescription(def, indentation = "", firstInBlock = !0) {
  const {description} = def;
  if (null == description) {
    return "";
  }
  return (indentation && !firstInBlock ? "\n" + indentation : indentation) + print({
    kind: Kind.STRING,
    value: description,
    block: isPrintableAsBlockString()
  }).replace(/\n/g, "\n" + indentation) + "\n";
}

export { printIntrospectionSchema, printSchema, printType };
